# -*- coding: utf-8 -*-
import os
import time
from datetime import datetime

import requests

from ks3.connection import Connection, PathCallingFormat, VirtualHostCallingFormat
from ks3.exception import S3ResponseError, S3CreateError
from ks3.key import Key
from ks3.tagging import Tag
from ks3.bucket import BucketLocation
from ks3.xmlParsers.bucketCors import BucketCors, CORSRule
from ks3.xmlParsers.bucketCrossReplicate import BucketCrossReplicate
from ks3.xmlParsers.bucketInventory import BucketInventory, Destination, Schedule, Filter as InventoryFilter
from ks3.xmlParsers.bucketLifecycle import BucketLifecycle, Rule as LifecycleRule, Filter as LifecycleFilter, \
  Expiration as LifecycleExpiration, Transition as LifecycleTransition, AbortIncompleteMultipartUpload
from ks3.xmlParsers.bucketLogging import BucketLogging
# 首先初始化AccessKeyId、AccessKeySecret、Endpoint等信息。
# 通过环境变量获取，或者把诸如“<你的AccessKeyId>”替换成真实的AccessKeyId等。
from ks3.xmlParsers.bucketMirror import BucketMirror, AsyncMirrorRule, SyncMirrorRules, MirrorRequestSetting, \
  HeaderSetting
from ks3.xmlParsers.bucketRetention import BucketRetention, Rule as RetentionRule
from base64 import urlsafe_b64encode
import asyncio
import ks3
import logging
import concurrent.futures
ks3.set_stream_logger(level=logging.DEBUG)

ak = os.getenv('KS3_IAM_ACCESS_KEY_ID', '<YOUR_ACCESS_KEY>')
sk = os.getenv('KS3_IAM_ACCESS_KEY_SECRET', '<YOUR_SECRET_KEY>')
bucket_name = os.getenv('KS3_BUCKET', '<KS3_TEST_BUCKET>')
endpoint = 'ks3-cn-beijing-internal.ksyuncs.com' #os.getenv('KS3_TEST_ENDPOINT', 'ks3-cn-shanghai-internal.ksyuncs.com')

conn = Connection(ak, sk, host=endpoint, port=80, ua_addon='test-ua/1', is_secure=False,
                  calling_format=VirtualHostCallingFormat(), timeout=(1, 1), retry_policy=None, dns_cache_time=0,
                  connection_pool_size=0) #port=8091,
key_name = 'test_key'


def getAllBuckets(project_ids=None):
  try:
    buckets = conn.get_all_buckets(project_ids=project_ids) #
    for b in buckets:
      print(b.name, b.region, b.Region, b.creation_date)
  except Exception as e:
    print('request_id: %s' % e.request_id)
    print(e)
    raise

def headBucket(bucket_name):
  # 如果正常返回，则Bucket存在；如果抛出S3ResponseError
  try:
    headResult = conn.head_bucket(bucket_name)
  except S3ResponseError as e:
    print('status: %s, reason: %s, request_id: %s, body: %s' % (e.status, e.reason, e.request_id, e.body))

def head_bucket_shadowcopy(bucket_name):
  # 如果正常返回，则Bucket存在；如果抛出S3ResponseError
  try:
    headResult = conn.head_bucket(bucket_name)
    if headResult.headers['x-kss-shadowcopy']:
      print('已开启shadowcopy', headResult.headers['x-kss-shadowcopy'])
    else:
      print('未开启shadowcopy')
  except S3ResponseError as e:
    print('status: %s, reason: %s, request_id: %s, body: %s' % (e.status, e.reason, e.request_id, e.body))

def getBucketLocation(bucket_name):
  loc = conn.get_bucket_location(bucket_name)
  print(isinstance(loc, BucketLocation))
  print(loc.location)

def createBucket(bucket_name, location=''):
  try:
    resp = conn.create_bucket(bucket_name, location=location, headers={
      'x-kss-bucket-type': 'ARCHIVE'
    })
    print('createBucket, request_id: ', resp.response_metadata.request_id)
  except Exception as e:
    print('request_id: %s' % e.request_id)
    if isinstance(e, S3CreateError):
      print("create bucket error, bucket already exists, error: ", e)
    else:
      print("create bucket error, error: ", e)

def deleteBucket(bucket_name):
  try:
    resp = conn.delete_bucket(bucket_name)
    print('deleteBucket, request_id: ', resp.response_metadata.request_id)
  except S3ResponseError as error:
    print('error: %s' % error)
    print('error requestid: %s' % error.request_id)


def getBucketAcl(bucket_name):
  b = conn.get_bucket(bucket_name)
  policy = b.get_acl()
  print("got acl_policy, request_id: ", policy.response_metadata.request_id)
  for grant in policy.acl.grants:
    print(grant.permission, grant.display_name, grant.email_address, grant.id)

def setBucketAcl(bucket_name):
  b = conn.get_bucket(bucket_name)
  resp = b.set_acl("private")
  print("set_acl, request_id: ", resp.response_metadata.request_id)

def manageBucketPolicy(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  setResp = bucket.set_bucket_policy(
    policy='{"Statement":[{"Resource":["krn:ksc:ks3:::test-bucket","krn:ksc:ks3:::test-bucket/*"],"Principal":{"KSC":["krn:ksc:iam::32432423:root"]},"Action":["ks3:*"],"Effect":"Allow"}]}')
  print("got setResp, request_id: ", setResp.response_metadata.request_id)
  policy = bucket.get_bucket_policy()
  print("got policy, request_id: ", policy.response_metadata.request_id)
  print(policy.data)
  deleteResp = bucket.delete_bucket_policy()
  print("got deleteResp, request_id: ", deleteResp.response_metadata.request_id)

def getBucketLifeCycle(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  lifecycle = bucket.get_bucket_lifecycle()
  print("get_bucket_lifecycle, request_id: ", lifecycle.response_metadata.request_id)
  print(lifecycle.to_xml())

def setBucketLifeCycle(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  lifecycle = BucketLifecycle()
  # id 和 status 必须
  rule = LifecycleRule(id='rule1', status='Enabled')
  date = datetime(2021, 9, 12).strftime('%Y-%m-%dT%H:%M:%S') + '+08:00'
  rule.expiration = LifecycleExpiration(date=date)
  lifecycle.rule = [rule]
  resp = bucket.set_bucket_lifecycle(lifecycle)
  print("set_bucket_lifecycle, request_id: ", resp.response_metadata.request_id)


def setBucketLifeCycle2(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  tag = Tag(key="tag1", value="test1")
  # 筛选前缀为 prefix1、标签为 tag1:test1 的 objects，设置过期规则，在其最后修改时间3天后过期。
  rule1 = LifecycleRule('rule1', LifecycleFilter('prefix1', tags=[tag]),
                        status='Enabled',
                        expiration=LifecycleExpiration(days=3))

  # 设置过期规则，筛选标签为 tag1:test1 的 objects，最后修改时间在指定日期之前的，过期
  rule2 = LifecycleRule('rule2', LifecycleFilter(tags=[tag]),
                        status='Enabled',
                        expiration=LifecycleExpiration(
                          date=datetime(2021, 12, 12).isoformat(timespec='seconds') + '+08:00'))

  # 设置存储类型转换规则，筛选前缀为 prefix3 的 objects，在其最后修改时间20天之后转为低频访问类型，在其最后修改时间30天之后转为归档类型。
  rule3 = LifecycleRule('rule3', LifecycleFilter('prefix3'),
                        status='Enabled',
                        transitions=[LifecycleTransition(days=20, storage_class='STANDARD_IA'),
                                     LifecycleTransition(days=60, storage_class='ARCHIVE')])

  # 设置存储类型转换规则，筛选前缀为 prefix3 的 objects，最后修改时间在指定日期之前的，转为低频访问类型
  rule4 = LifecycleRule('rule4', LifecycleFilter('prefix4'),
                        status='Enabled',
                        transitions=[
                          LifecycleTransition(date=datetime(2021, 12, 12).isoformat(timespec='seconds') + '+08:00',
                                              storage_class='STANDARD_IA')])
  # 设置分片上传碎片清理规则，筛选前缀为 prefix5 的 part，在最后修改时间7天后删除
  rule5 = LifecycleRule('rule5', LifecycleFilter('prefix5'),
                        status='Enabled',
                        abort_incomplete_multipart_upload=AbortIncompleteMultipartUpload(days_after_initiation=7))

  # print(rule1.to_xml())
  # print(rule2.to_xml())
  # print(rule3.to_xml())
  # print(rule4.to_xml())
  # print(rule5.to_xml())
  lifecycle = BucketLifecycle([rule1, rule2, rule3, rule4, rule5])
  resp = bucket.set_bucket_lifecycle(lifecycle)
  print("set lifecycle, request_id: ", resp.response_metadata.request_id)

def deleteBucketLifeCycle(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.delete_bucket_lifecycle()
  print("delete_bucket_lifecycle, request_id: ", resp.response_metadata.request_id)

def enableBucketLogging(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.enable_logging(bucket, target_prefix='hehehehe')
  print("enable_logging, request_id: ", resp.response_metadata.request_id)

def disableBucketLogging(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.disable_logging()
  print("disable_logging, request_id: ", resp.response_metadata.request_id)

def getBucketCors(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  cors = bucket.get_bucket_cors()
  print("got cors, request_id: ", cors.response_metadata.request_id)
  print(cors.to_xml())

def putBucketCors(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  cors = BucketCors([CORSRule(origins=["http://dev.ksyun.com"], methods=["GET", "HEAD"], max_age="200", headers=["content-type"], exposed_headers=["content-type", "x-kss-acl"])])
  print('cors: ', cors.to_xml())
  resp = bucket.set_bucket_cors(cors)
  print("set_bucket_cors, request_id: ", resp.response_metadata.request_id)

def deleteBucketCors(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.delete_bucket_cors()
  print("delete_bucket_cors, request_id: ", resp.response_metadata.request_id)

def getBucketCrr(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  crr = bucket.get_bucket_crr()
  print("got bucket crr, request_id: ", crr.response_metadata.request_id)
  print(crr.to_xml())

def setBucketCrr(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  # deleteMarkerStatus 的值为 Enabled 和 Disabled
  # historicalObjectReplication 表示是否开启存量复制，值为 Enabled 和 Disabled
  resp = bucket.set_bucket_crr('test-bucket-repli', deleteMarkerStatus=BucketCrossReplicate.ENABLED,
                        historicalObjectReplication=BucketCrossReplicate.ENABLED, prefix=['hello'])
  print("set bucket crr, request_id: ", resp.response_metadata.request_id)

def deleteBucketCrr(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.delete_bucket_crr()
  print("delete bucket crr, request_id: ", resp.response_metadata.request_id)

def getBucketLogging(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.get_bucket_logging()
  print("got bucket logging, request_id: ", resp.response_metadata.request_id)
  print("got bucket logging, xml: ", resp.to_xml())

def setBucketLogging(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  blogging = BucketLogging(target=bucket_name)
  resp = bucket.set_bucket_logging(blogging.to_xml())
  print("set_bucket_logging, request_id: ", resp.response_metadata.request_id)

def getBucketMirror(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  mirror = bucket.get_bucket_mirror()
  print("got bucket mirror, request_id: ", mirror.response_metadata.request_id)
  print("got bucket mirror, content: ", mirror.data)

def setBucketMirror(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  set_headers = [{
      "key": "d",
      "value": "b"
  }]
  remove_headers = [{
      "key": "d"
  }]
  pass_headers = [{
      "key": "abc"
  }]
  header_setting = HeaderSetting(set_headers=set_headers, remove_headers=remove_headers, pass_all=False, pass_headers=pass_headers)
  mirror_request_setting = MirrorRequestSetting(pass_query_string=False, follow3xx=False, header_setting=header_setting)
  async_mirror_rule = AsyncMirrorRule.rule_with_acl(mirror_urls=["http://abc.om", "http://www.wps.cn"], saving_setting_acl="private")
  sync_mirror_rules = SyncMirrorRules.rules_with_prefix_acl(key_prefixes=["abc"], mirror_url="http://v-ks-a-i.originalvod.com", mirror_request_setting=mirror_request_setting, saving_setting_acl="private")
  mirror = BucketMirror(use_default_robots=False, async_mirror_rule=async_mirror_rule, sync_mirror_rules=[sync_mirror_rules])
  resp = bucket.set_bucket_mirror(mirror)
  print("set bucket mirror, request_id: ", resp.response_metadata.request_id)

def deleteBucketMirror(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.delete_bucket_mirror()
  print("delete bucket mirror, request_id: ", resp.response_metadata.request_id)


def setBucketRetention(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  bucket_retention = BucketRetention()
  # 设置回收站规则：启用/禁用、文件保留天数
  rule = RetentionRule(RetentionRule.ENABLED, 2)
  # rule = RetentionRule(RetentionRule.DISABLED, 2)
  bucket_retention.rule = rule
  bucket.set_bucket_retention(bucket_retention)


def getBucketRetention(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  ret = bucket.get_bucket_retention()
  print(ret.to_xml())


def listRetention(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  keys = bucket.list_retention()
  for k in keys:
    print(k.name)
    print(k.retention_id)
    print(k.storage_class)


def clearObject(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.new_key('clear_key_name')
  k.clear_object('<retention_id>')


def recoverObject(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.new_key('recover_key_name')
  # 如果retention_id为空，默认恢复回收站中指定key的最新版本
  k.recover_object(overwrite=True, retention_id='<retention_id>')


def setBucketInventory(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  # 添加清单规则，规则名为inventory1，状态为开启
  # 清单报告以CSV的格式存储，存储路径为target_account账号下的target_bucket桶，前缀为inventory1_的文件
  # 报告生成频率为每周一次
  # 扫描指定前缀和最后修改时间范围的对象
  # 清单结果中包含 文件大小 和 最后修改时间 的信息
  inventory = BucketInventory(id='inventory1', is_enabled=True,
                              destination=Destination(Destination.FORMAT_CSV,
                                                      bucket=bucket_name, prefix='inventory1_'),
                              schedule=Schedule(Schedule.WEEKLY),
                              filter=InventoryFilter(prefix='a', last_modify_begin_time_stamp=int(time.time())),
                              optional_fields=[BucketInventory.OPTIONAL_FIELD_SIZE,
                                               BucketInventory.OPTIONAL_FIELD_LAST_MODIFIED_DATE])
  print(inventory.to_xml())
  resp = bucket.set_bucket_inventory(inventory)
  print("set_bucket_inventory, request_id: ", resp.response_metadata.request_id)

def getBucketInventory(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  inventory = bucket.get_bucket_inventory('inventory1')
  print("get_bucket_inventory, request_id: ", inventory.response_metadata.request_id)
  print(inventory.to_xml())

def listBucketInventory(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  result = bucket.list_bucket_inventory()
  print("list_bucket_inventory, request_id: ", result.response_metadata.request_id)
  for inventory in result.inventory_configurations:
    print(inventory.to_xml())

def deleteBucketInventory(bucket_name):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.delete_bucket_inventory('inventory1')
  print("delete_bucket_inventory, request_id: ", resp.response_metadata.request_id)


#####################  ks3.billing  ###########################
from ks3.billing import get_buckets_data, query_ks3_data, query_bucket_rank

def getBucketsData(bucket_names=None):
  try:
    data = get_buckets_data(ak, sk, action="QueryBucketRank", start_time="202311192300", end_time="202311192359", bucket_names=bucket_names, products="", inner_api=True)
    print(data)
  except Exception as e:
    print(e.request_id)
    print(e)

def queryKs3Data(bucket_names=None):
  try:
    data = query_ks3_data(ak, sk, start_time="202311192300", end_time="202311192359", bucket_names=bucket_names, inner_api=True)
    print(data)
  except Exception as e:
    print(e.request_id)
    print(e)

def queryBucketRank(number):
  try:
    data = query_bucket_rank(ak, sk, start_time="202406192300", end_time="202407192359", inner_api=True, number=number)
    print(data)
  except Exception as e:
    print(e.request_id)
    print(e)

#####################  ks3.object  ###########################
def get_object_meta(bucket_name, object_key_name, headers=None):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.get_key_meta(object_key_name, headers=headers)
  if resp:
    print("获取文件header成功: \n", resp.data.headers)
    print('get_key_meta, request_id: %s' % resp.response_metadata.request_id)

def uploadObjectFromFile(filename):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.new_key(filename)
  ret = k.set_contents_from_filename(os.path.expanduser("~") + '/Downloads/' + filename)
  if ret and ret.status == 200:
    print("上传成功")
    print("requestid:", ret.headers['x-kss-request-id'])
    print("requestid:", ret.response_metadata.request_id)

# def upload_async(filename):
#   bucket = conn.get_bucket(bucket_name)
#   k = bucket.new_key(filename)
#   ret = asyncio.run(k.upload_file_async(os.path.expanduser("~") + '/Downloads/' + filename))
#   if ret and ret.status == 200:
#     print("上传成功")

def uploadObjectFromString(type='ARCHIVE'):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.new_key('大家好')
  # # key 和 value 需要 url 编码
  # taggingStr = 'name=jh'
  # headers = {'x-kss-tagging': taggingStr}
  headers = {}
  if type is not None:
    headers['x-kss-storage-class'] = type

  ret = k.set_contents_from_string('1234', headers=headers)
  # # 请求ID。请求ID是本次请求的唯一标识，强烈建议在程序日志中添加此参数。
  # print(ret.headers['x-kss-request-id'])
  print(ret.response_metadata.request_id)
  # # ETag是put_object方法返回值特有的属性，用于标识一个Object的内容。
  # print(ret.headers)
  # HTTP返回码。
  if ret and ret.status == 200:
    print("上传成功")

def headObject():
  bucket = conn.get_bucket(bucket_name)
  k = bucket.get_key('test_encryption')
  if k:
    print(k.name, k.size, k.last_modified, k.object_type, k.tagging_count)

def downloadObjectAndPrint(keyname, byte_range=(0, 1), headers=None):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.get_key(keyname)
  ret = k.get_contents_as_string(byte_range=byte_range, headers=headers)
  print('request_id: %s' % ret.response_metadata.request_id)
  s = ret.data.decode()
  print('打印字符串: %s' % s)

# 定义一个模拟 I/O 操作的协程
async def io_task(task_id, delay):
    print(f"Task {task_id} started, will sleep for {delay} seconds")
    await asyncio.sleep(delay)  # 模拟 I/O 操作
    print(f"Task {task_id} finished")

async def _wrap_func(method, keyname_or_filename):
    start = datetime.now()
    result = await method(keyname_or_filename)
    end = datetime.now()
    return result, start, end
async def download_async(keynames, headers=None):
  async_start_time = datetime.now()
  bucket = conn.get_bucket(bucket_name)
  tasks = []
  for index, keyname in enumerate(keynames):
    k = bucket.get_key(keyname)
    # task = _wrap_func(k.download_file_async, os.path.expanduser("~") + '/Downloads/' + keyname)
    print('download_async[%s]: %s' % (async_start_time, os.path.expanduser("~") + '/Downloads/down_' + keyname))
    task = k.download_file_async(os.path.expanduser("~") + '/Downloads/down_' + keyname)
    tasks.append(task)
  rets = await asyncio.gather(*tasks, return_exceptions=True)
  # 打印某一个任务的requestid
  print('rets[0] requestid:', rets[0].response_metadata.request_id)
  async_end_time = datetime.now()
  # print('下载异步耗时: ', async_end_time - async_start_time)
  return rets

def upload_files(keynames, headers=None):
  bucket = conn.get_bucket(bucket_name)
  tasks = []
  for keyname in keynames:
    k = bucket.new_key(keyname)
    print('upload_files', os.path.expanduser("~") + '/Downloads/' + keyname)
    k.set_contents_from_filename(os.path.expanduser("~") + '/Downloads/' + keyname, headers=headers)

async def upload_async(filenames, headers=None):
  async_start_time = datetime.now()
  bucket = conn.get_bucket(bucket_name)
  tasks = []
  for index, filename in enumerate(filenames):
    k = bucket.new_key(filename)
    print('upload_async[%s]: %s' % (async_start_time, os.path.expanduser("~") + '/Downloads/' + filename))
    task = k.upload_file_async(os.path.expanduser("~") + '/Downloads/' + filename)
    tasks.append(task)
  results = await asyncio.gather(*tasks, return_exceptions=True)
  # 打印某一个任务的requestid
  print('results[0] requestid:', results[0].response_metadata.request_id)
  return results

  # 打印结果
  for i, result in enumerate(results, 1):
      if isinstance(result, Exception):
          print(f"Task {i} raised an exception: {result}")
      else:
        if result and result.status == 200:
          print(f"Task {i} 上传成功")
  async_end_time = datetime.now()
  # print('上传异步耗时: ', async_end_time - async_start_time)

def download_file(keyname, headers=None):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.get_key(keyname)
  print('download_file', os.path.expanduser("~") + '/Downloads/' + keyname)
  ret = k.get_contents_to_filename(os.path.expanduser("~") + '/Downloads/' + keyname, headers=headers)
  print('download_file request_id: %s' % ret.response_metadata.request_id)

def download_files(keynames, headers=None):
  bucket = conn.get_bucket(bucket_name)
  tasks = []
  for keyname in keynames:
    k = bucket.get_key(keyname)
    print('download_files', os.path.expanduser("~") + '/Downloads/down_' + keyname)
    k.get_contents_to_filename(os.path.expanduser("~") + '/Downloads/down_' + keyname, headers=headers)

def downloadObjectAsStreamAndPrint():
  bucket = conn.get_bucket(bucket_name)
  k = bucket.get_key('shake.txt')
  bytes = k.read(300)
  print('start: ', datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"))
  while bytes:
    s = bytes.decode()
    print('bytes decoded:', s)
    time.sleep(1)
    bytes = k.read(300)
  print('end: ', datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"))

def downloadObjectAndSave(key_name, byte_range=(0, 1), headers=None):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.get_key(key_name)
  k.get_contents_to_filename('/Users/jabbar/Downloads/DSC03380.encrypt.非分块.jpg.download', headers=headers)

def deleteObject(key_name):
  bucket = conn.get_bucket(bucket_name)
  try:
    ret = bucket.delete_key(key_name)
    print("删除成功, request_id: ", ret.response_metadata.request_id)
  except Exception as e:
    print("删除失败")
    print(e)

def getObjectAcl():
  bucket = conn.get_bucket(bucket_name)
  policy = bucket.get_acl('article.txt')
  print("got acl_policy, request_id: ", policy.response_metadata.request_id)
  for grant in policy.acl.grants:
    print(grant.permission, grant.display_name, grant.email_address, grant.id)
  print(policy.to_xml())

def setObjectAcl():
  bucket = conn.get_bucket(bucket_name)
  # object policy : private | public-read | public-read-write
  bucket.set_acl("public-read", '<YOUR_KEY_NAME>')

def setObjectMeta():
  b = conn.get_bucket(bucket_name)
  b.copy_key('<yourKeyName>', '<yourBucketName>', '<yourKeyName>',
             headers={'content-type': 'text/plain', 'x-kss-metadata-directive': 'REPLACE'})

def setObjectStorageClass():
  b = conn.get_bucket(bucket_name)
  b.copy_key('<yourKeyName>', '<yourBucketName>', '<yourKeyName>', headers={'x-kss-storage-class': 'STANDARD_IA'})

def copy(dstKey, srcKey, dst_bucket_name=None, headers=None):
  # bucket_name = 'happyhour'
  b = conn.get_bucket(bucket_name)
  ret = b.copy_key(dstKey, dst_bucket_name, srcKey, headers=headers)
  print("copy_key, request_id: ", ret.response_metadata.request_id)
  return ret

def copy_encryption(dstKey, srcKey, encrypt_key=None):
  b = conn.get_bucket(bucket_name)
  b.copy_key(dstKey, bucket_name, srcKey, encrypt_key=encrypt_key)

def list_objects_truncated():
  bucket = conn.get_bucket('test-bucket')
  keys = bucket.get_all_keys(max_keys=3)
  print("got keys: request_id is: ", keys.response_metadata["request_id"])
  for k in keys:
    print('object:', k.name)

def list_objects():
  bucket = conn.get_bucket('test-bucket')
  keys = bucket.list(prefix='16')
  print("list, request_id: ", keys.response_metadata.request_id)
  res1 = [k.name for k in keys]
  print('res1 : ', res1)

  keys2 = bucket.listObjects(prefix='16')
  # res2 = [k.name for k in keys2]
  print('keys2 : ', keys2)
  # for k in keys:
  #   print('object:', k.name)

def list_objects_v2_truncated():
  bucket = conn.get_bucket('test-bucket')
  keys = bucket.get_all_keys(list_type=2, max_keys=3)
  print("got keys: request_id is: ", keys.response_metadata["request_id"])
  for k in keys:
    print('object:', k.name)

def list_objects_v2(delimiter='#', prefix=None, max_keys=None, marker=None, encoding_type='', fetch_owner=True):
  bucket = conn.get_bucket(bucket_name)
  keys = bucket.list_v2(delimiter=delimiter, prefix=prefix, max_keys=max_keys, marker=marker, encoding_type=encoding_type, fetch_owner=fetch_owner)
  for k in keys:
    print('object:', k.name)

def list_objects_v2_no_params():
  bucket = conn.get_bucket(bucket_name)
  keys = bucket.list_v2()
  for k in keys:
    print('object:', k.name)


def listObjectsMore(bucket_name, delimiter=None, prefix=None, max_keys=None, marker=None):
  bucket = conn.get_bucket(bucket_name)
  keys = bucket.list(delimiter=delimiter, prefix=prefix, max_keys=max_keys, marker=marker)
  # print(len(keys))
  res = [k.name for k in keys]
  print('res: ', res)
  # print('item:', k.name, type(k))

def listAndDelete():
  bucket = conn.get_bucket(bucket_name)
  keys = bucket.list(delimiter='/', max_keys=10, prefix='15')
  print(keys.marker)
  for k in keys:
    print('object:', k.name)
    deleteObject(k.name)

def listObjectsAndFilter(endTime=None):
  bucket = conn.get_bucket('auto-test-bucket')
  keys = bucket.listObjects(delimiter='/', max_keys=2) # start_time=1640331446, end_time=1641895096
  for k in keys:
    print(k)

def getObjectTagging():
  bucket = conn.get_bucket(bucket_name)
  key = bucket.get_key('testTagging')
  tagging = key.get_object_tagging()
  print("get_object_tagging, request_id: ", tagging.response_metadata.request_id)
  print(tagging.to_xml())

def setObjectTagging():
  bucket = conn.get_bucket(bucket_name)
  key = bucket.get_key('testTagging')
  tagging = [Tag('0'), Tag('1', '1')]
  ret = key.set_object_tagging(tagging)
  print("set_object_tagging, request_id: ", ret.response_metadata.request_id)

def deleteObjectTagging():
  bucket = conn.get_bucket(bucket_name)
  key = bucket.get_key('testTagging')
  ret = key.delete_object_tagging()
  print("delete_object_tagging, request_id: ", key.response_metadata.request_id)

def calcFolderSize(bucket, folder):
  length = 0
  keys = bucket.list(prefix=folder)
  for k in keys:
    if isinstance(k, Key):
      length += k.size
  return length

from ks3.prefix import Prefix

# 列举指定目录下的文件大小
def getFolderSizeInBucket():
  bucket = conn.get_bucket(bucket_name)
  keys = bucket.list(delimiter='/')
  for k in keys:
    if isinstance(k, Prefix):
      print('dir: ' + k.name + '  size:' + str(calcFolderSize(bucket, k.name)) + "Byte")


def test_multipart_upload(file_path = None, encrypt_key=None):
  import math, os
  from filechunkio import FileChunkIO
  bucket = conn.get_bucket(bucket_name)

  source_path = '/Users/jiahua/Downloads/DSC03380.JPG' if not file_path else file_path
  # 源文件大小
  source_size = os.stat(source_path).st_size

  mp = bucket.initiate_multipart_upload("DSC03380.mp.jpg", encrypt_key=encrypt_key)
  print(mp)
  print('initiate_multipart_upload, requestid: ', mp.response_metadata.request_id)

  chunk_size = 5242880
  chunk_count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))

  # 通过 FileChunkIO 将文件分片
  for i in range(chunk_count):
    offset = chunk_size * i
    bytes = min(chunk_size, source_size - offset)
    with FileChunkIO(source_path, 'r', offset=offset, bytes=bytes) as fp:
      # 逐个上传分片
      ret = mp.upload_part_from_file(fp, part_num=i + 1)
      print('upload_part_from_file, requestid: ', ret.response_metadata.request_id)
  print(mp.to_xml())
  # 发送请求，合并分片，完成分片上传
  cmp = mp.complete_upload()
  print('complete_upload, requestid: ', cmp.response_metadata.request_id)
  if cmp and cmp.status == 200:
    print("上传成功")

def test_fetch_object():
  bucket = conn.get_bucket(bucket_name)
  key = bucket.new_key('www-logo')
  ret = key.fetch_object(source_url='http://fe.ksyun.com/project/resource/img/www-logo.png',
                      headers={'x-kss-acl': 'public-read'})
  print(ret.headers['x-kss-request-id'])
  if ret and ret.status == 200:
    print('fetch成功')

def test_generate_url(key_name, image_attrs=None):
  b = conn.get_bucket(bucket_name)
  k = b.get_key(key_name)
  if k:
    url = k.generate_url(600, image_attrs=image_attrs)  # 60s 后该链接过期
    print(url)
  else:
    print('object not found')

def test_get_presigned_url(method=None, response_headers=None, key_name=None):
  b = conn.get_bucket(bucket_name)
  url = b.get_presigned_url(600, method=method, response_headers=response_headers, key_name=key_name)  # 60s 后该链接过期
  print(url)
  return url

def restoreObject(key_name):
  b = conn.get_bucket(bucket_name)
  k = b.get_key(key_name)
  ret = k.restore_object()
  print('restore_object, request_id: ', ret.response_metadata.request_id)

def test_put_via_presigned_url(key):
  url = test_get_presigned_url(key)
  with open('./article.txt', 'rb') as fp:
    result = requests.put(url, data=fp)
    print(result)

from ks3.sts import assumeRole
def test_assumeRole():
  try:
    print(assumeRole(ak, sk, "krn:ksc:iam::xxx:role/xx-test-bucket", "ks3", 3600))
  except Exception as e:
    print(e.request_id)
    print(e)

def put_object_callback(key_name):
  b = conn.get_bucket('hanjing-test000')
  # 新建对象key
  k = b.new_key(key_name)
  # 上传成功后，向回调地址POST数据
  # 需要设置`x-kss-callbackurl`和`x-kss-callbackbody`请求头，详见文档[上传回调处理](https://docs.ksyun.com/documents/956)。
  # x-kss-callbackurl 为发起回调时请求的服务器地址
  # x-kss-callbackbody 为发起回调时请求的body的值
  headers = {"x-kss-callbackurl": "https://yh-sh.ks3-cn-shanghai.ksyuncs.com", "x-kss-callbackbody": "objectKey=${key}&etag=${etag}&bucket=${bucket}&objectSize=${objectSize}&mimeType=${mimeType}&createTime=${createTime}", "x-kss-callbackauth":1}
  ret = k.set_contents_from_string("test-callback", headers=headers)
  if ret and ret.status == 200:
    print("上传成功")

def append_object_from_file(key_name, filename, position=0):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.new_key(key_name)
  ret = k.append_object_from_filename('/Users/jiahua/Downloads/' + filename, position=position)
  if ret and ret.status == 200:
    print("上传成功")


def append_object_from_string(key_name, content, position=0):
  bucket = conn.get_bucket(bucket_name)
  k = bucket.new_key(key_name)
  ret = k.append_object_from_string(content, position=position)
  print("x-kss-next-append-position:", ret.headers['x-kss-next-append-position'])
  print("requestid:", ret.response_metadata.request_id)
  if ret and ret.status == 200:
    print("上传成功")
  return ret

def getObjectMeta(object_key_name, headers=None):
  bucket = conn.get_bucket(bucket_name)
  resp = bucket.get_key_meta(object_key_name, headers=headers)
  if resp:
    print("获取文件header成功: ", resp.headers)
    print("next-append: ", type(resp.headers['x-kss-next-append-position']))
  return resp

async def up_down():
  async_start_time = datetime.now()
  ret = await asyncio.gather(
    download_async(
      ['DSC03380.JPG']),
    upload_async(
      ['DSC03380.JPG'])
  )
  # await download_async(['DSC03380.JPG', 'DSC03380.cp.jpg'])
  async_end_time = datetime.now()
  print('异步耗时: ', async_end_time - async_start_time)
  return ret

# try:
#   result = append_object_from_string("test_append", "hello ")
# except S3ResponseError as e:
#   print('status: %s, reason: %s, request_id: %s, body: %s' % (e.status, e.reason, e.request_id, e.body))

#
start = time.time()
try:
  result = getObjectMeta("test_append")
except:
  print(f'耗时: {time.time() - start} 秒')
  raise
# append_object_from_string("test_append3", "world", position=int(result.headers['x-kss-next-append-position']))
# append_object_from_file("test_append3", "test_append", position=int(result.headers['x-kss-next-append-position']))

# getAllBuckets()
# head_bucket_shadowcopy('a-beijing-bucket')
# getBucketLocation('test-host-style')
# createBucket('test-bucket-222222', location='SHANGHAI')
# deleteBucket('test-bucket-222222')
# getBucketAcl('test-bucket')
# setBucketAcl('test-bucket')
# try:
#   getBucketLifeCycle(bucket_name)
# except:
#   setBucketLifeCycle(bucket_name)

# setBucketLifeCycle2(bucket_name)
# getBucketLifeCycle(bucket_name)
# deleteBucketLifeCycle(bucket_name)
# getBucketLogging(bucket_name)
# setBucketLogging(bucket_name)
# disableBucketLogging(bucket_name)
# enableBucketLogging(bucket_name)
# putBucketCors(bucket_name)
# getBucketCors(bucket_name)
# deleteBucketCors(bucket_name)
# setBucketCrr(bucket_name)
# getBucketCrr(bucket_name)
# deleteBucketCrr(bucket_name)
# setBucketMirror(bucket_name)
# getBucketMirror(bucket_name)
# deleteBucketMirror(bucket_name)
# manageBucketPolicy(bucket_name)
# getBucketMirror(bucket_name)
# setBucketInventory(bucket_name)
# getBucketInventory(bucket_name)
# listBucketInventory(bucket_name)
# deleteBucketInventory(bucket_name)
##### object #####
# get_object_meta(bucket_name, 'article.txt')
# list_objects_truncated()
# list_objects()
# list_objects_v2_truncated()
# list_objects_v2(delimiter=None)
# list_objects_v2_no_params()
# listObjectsMore("auto-test-bucket", delimiter='/', max_keys=2, marker='1050')
# listObjectsMore("auto-test-bucket", delimiter='/', max_keys=2, marker='105/')
# listObjectsMore("auto-test-bucket", delimiter='/', max_keys=2, marker='105%2F')
# listObjectsAndFilter()
# setObjectTagging()
# getObjectTagging()
# deleteObjectTagging()
# getFolderSizeInBucket();
# downloadObjectAndPrint(keyname="1659684872.torrent")
# downloadObjectAsStreamAndPrint()
# downloadObjectAndSave(bucket_name, 'DSC03380.encrypt.非分块.jpg', headers={
#   'range': 'bytes=0-5242880'
# })
# count = 20
# while True:
# uploadObjectFromFile('3_7240456.jpeg')

# 异步下载和同步下载对比

#
# try:
#   # 'DSC03380.cp.jpg', 'DSC03380.mp.jpg', 'DSC03384.JPG', 'DSC03384.cp.JPG', 'DSC03384.cp2.JPG'
# run_ret = asyncio.run(up_down())
# print(run_ret)

# download_file('DSC03380.JPG')
#   # asyncio.run(upload_async(
#   #   ['DSC03380.JPG', 'DSC03380.jpg', 'DSC03380.cp.jpg', 'DSC03380.mp.jpg', 'DSC03384.JPG', 'DSC03384.cp.JPG',
#   #    'DSC03384.cp2.JPG']))
#   # print(download_results[0].request_id)
#   # print(download_results[1].request_id)
#   # print(download_results[2].request_id)
#   # start_time = datetime.now()
#   # download_files(['DSC03380.JPG', 'DSC03380.jpg', 'DSC03380.cp.jpg', 'DSC03380.mp.jpg', 'DSC03384.JPG', 'DSC03384.cp.JPG'])
#   # upload_files(
#   #   ['DSC03380.JPG', 'DSC03380.jpg', 'DSC03380.cp.jpg', 'DSC03380.mp.jpg', 'DSC03384.JPG', 'DSC03384.cp.JPG',
#   #    'DSC03384.cp2.JPG'])
#   # end_time = datetime.now()
#   # print('同步耗时: ', end_time - start_time)
# except Exception as e:
#   print('下载失败', e)
#   raise

# # 异步上传和同步上传对比
# try:
#   asyncio.run(upload_async(
#     ['DSC03380.JPG', 'DSC03380.jpg', 'DSC03380.cp.jpg', 'DSC03380.mp.jpg', 'DSC03384.JPG', 'DSC03384.cp.JPG',
#      'DSC03384.cp2.JPG']))
#
#   start_time = datetime.now()
#   upload_files(
#     ['DSC03380.JPG', 'DSC03380.jpg', 'DSC03380.cp.jpg', 'DSC03380.mp.jpg', 'DSC03384.JPG', 'DSC03384.cp.JPG',
#      'DSC03384.cp2.JPG'])
#   end_time = datetime.now()
#   print('上传同步耗时: ', end_time - start_time)
# except Exception as e:
#   print('上传失败', e)

# upload_files(['DSC03380.JPG'])

# uploadObjectFromFile('xab')
# uploadObjectFromFile('法治.jpeg')
# upload_async('11.jpeg')
# uploadObjectFromFile('11.jpeg')
# uploadObjectFromFile('xac')
#   time.sleep(5)
#   count = count - 1
# uploadObjectFromString()
# headObject()
# downloadObjectAndPrint('test_directory/')
# deleteObject("大家好")
# getObjectAcl()

# test_multipart_upload()

# test_fetch_object()
# def test_generate_url(key_name, image_attrs=None):
#   b = conn.get_bucket(bucket_name)
#   k = b.get_key(key_name)
#   if k:
#     url = k.generate_url(600, image_attrs=image_attrs)  # 60s 后该链接过期
#     print(url)
#   else:
#     print('object not found')

# bucket = conn.get_bucket('test-bucket')
# keys = bucket.list()
# for k in keys:
#   print(k)
#   if isinstance(k, Key):
#     url = k.generate_url(600)
#     res = requests.get(url)
#     if res.status_code != 200:
#       print(url)


# res = requests.get(url)
# assert res.status_code != 403
# test_generate_url('\\a\\\\:<>*?|:q.jpeg')
# test_generate_url('#x2;.jp.csv')
# test_generate_url('10086/198.18-reset-共计9条.txt')
# test_generate_url('10086/getbucket.csv')
# test_put_via_presigned_url('index.html')
# list_objects_v2(delimiter='/', encoding_type='url', fetch_owner=False)
# list_objects()
# listObjectsMore()
# deleteObject()
# listObjectsAndFilter()
# listAndDelete()

# restoreObject("大家好")
# getBucketsData("test-bucket")
# queryKs3Data("test-bucket")
# queryBucketRank(10)
# test_assumeRole()
# getBucketLogging(bucket_name)

# copy_encryption('article.encryption.txt', 'article.txt', encrypt_key=True)

# fileName = "\u006b\u0073\u005f\u0073\u0063\u0061\u006e\u005f\u006f\u0063\u0072\u002f\u0032\u0030\u0031\u0038\u002d\u0030\u0033\u002d\u0031\u0032\u002f\u0010\ufffd\u0070\ufffd\ufffd\ufffd\u0002\u002e\u0057\u007f\u005b\ufffd\u0008\ufffd\u002a\ufffd\u002e\u006a\u0070\u0067";
# hehe = "�p���.W[�*�.jpg"
# for c in hehe:
#     print(repr(c), c)
# urlcoded = urllib.parse.quote(fileName)
# unquoted = urllib.parse.unquote('%FF')
# quoted = urllib.parse.quote(unquoted)
# print(unquoted)
# print(quoted)

# getObjectMeta('test-bucket', 'article.txt')
# key = copy('article.txt', 'article.txt', dst_bucket_name='test-bucket')
# getObjectMeta('test-bucket', 'article.txt')
# getObjectMeta('test-bucket', 'DSC03380.ssec.jpg')
# uploadObjectFromFile('DSC03380.jpg')

# import base64
# import hmac
# import hashlib
# import urllib
# h = hmac.new("accesskey",
#              "GET\n\n\n1141889120\n/examplebucket/oss-api.pdf",
#              hashlib.sha1)
# urllib.quote (base64.encodestring(h.digest()).strip())

# b = conn.get_bucket(bucket_name)
# all_uploads = b.get_all_multipart_uploads()
# print("get_all_multipart_uploads, request_id: ", all_uploads.response_metadata.request_id)
# for p in all_uploads:
#   some_mp = p
#   ret = p.cancel_upload()
#   if ret and ret.status == 204:
#     print('cancel_upload success')
#   print('cancel_upload, id: %s, status: %s, requestid: %s' % (p.id, ret.status, ret.response_metadata.request_id))
#   break

# all_parts = some_mp.get_all_parts()
# print("get_all_parts, request_id: ", all_parts.response_metadata.request_id)
# for part in all_parts:
#   print('part etag:%s' % part.etag)
#   print('part_number:%s' % part.part_number)


# 获取指定分片上传
# p = b.get_all_multipart_uploads(key_marker='DSC03380.encrypt.jpg-dddddddd')[0]
# p.cancel_upload()

# all_uploads = b.get_all_multipart_uploads(key_marker='DSC03380.encrypt.jpg-dddddddd')#max_uploads=27)
# print(len(all_uploads))
# for p in all_uploads:
#   print(p, p.id)
# put_object_callback('test-callback')

# import math, os
#
# b = conn.get_bucket(bucket_name)
# # 目标key名
# target_key_name = 'test_copy_part'
# # 获取源key的元数据信息
# source_key = b.get_key("DSC03380.jpg")
# source_size = source_key.size
#
# mp = b.initiate_multipart_upload(target_key_name)
# print('initiate_multipart_upload, requestid: ', mp.response_metadata.request_id)
#
# chunk_size = 5242880
# count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))
#
# for i in range(count):
#     start = chunk_size * i
#     end = min(start + chunk_size - 1, source_size - 1)
#     ret = mp.copy_part_from_key(bucket_name, source_key.name, i + 1, start, end)
#     print(ret.etag, ret.last_modified, ret.ChecksumCRC64ECMA)
#     print('copy_part_from_key, requestid: ', ret.response_metadata.request_id)
#
# cmp = mp.complete_upload()
# print('complete_upload, requestid: ', cmp.response_metadata.request_id)
# if cmp and cmp.status == 200:
#     print("复制成功")

# 生成listObjects外链
# b = conn.get_bucket('happyhour')
# # 拉取prefix为16/的对象列表
# list_params = {'prefix': '16/'}
# # conditions = [['starts-with','$key', '16/'], {'bucket': bucket_name}]
# conditions = [
#   {"bucket": "gzz-test"},
#   ["starts-with", "$key", "abc/"],
#   ["starts-with", "$key", "aaa/abc/"],
#   ["eq", "$key", "exampleobject"],
#   ["eq", "$key", "exampleobject1"]
# ]
# # 外链地址60s后过期
# url = b.presign(1741746992, method="GET", params=list_params, conditions=conditions, expires_in_absolute=True)
# print(url)

# share_code = b.create_share_code(auth_code=b'123456', prefix="16/")
# print('share_code:', share_code)
#
# from ks3.share_encryption import ShareCryptor
# shareCryptor = ShareCryptor(b'123456')
# dec_url = shareCryptor.decrypt(share_code)
# print(dec_url)

# # 生成分块上传的外链
# init_mpu_url = b.get_presigned_url(60, method="POST", key_name='DSC03380.mp.jpg', params={'uploads': None})
# print(init_mpu_url)
# result = requests.post(url=init_mpu_url)
# print(result.headers['x-kss-request-id'])
# print(result)


# # list分块上传任务
# list_mpu_params = {'key-marker': 'DSC03380.mp.jpg', 'upload-id-marker': '!', 'uploads': None, 'max-uploads': '2'}
# list_mpu_url = b.get_presigned_url(60, method="GET", params=list_mpu_params)
# print(list_mpu_url)
# list_mpu_result = requests.get(url=list_mpu_url)
# # print(list_mpu_result.content)
# # xml序列化content，然后取最后一个
# from xml.etree import ElementTree
# root = ElementTree.fromstring(list_mpu_result.content)
# # 提取命名空间
# namespace = {'ns': 'http://s3.amazonaws.com/doc/2006-03-01/'}
# uploads = root.findall('.//ns:Upload', namespace)
# last_upload = uploads[0]
# key = last_upload.find('ns:Key', namespace).text
# upload_id = last_upload.find('ns:UploadId', namespace).text
# print(key, upload_id)
#
# upload_mpu_url = b.get_presigned_url(60, method="PUT", key_name=key, params={'partNumber': '1', 'uploadId': upload_id})
# print(upload_mpu_url)
# result = requests.put(url=upload_mpu_url, data='hello')
# print(result)
# print(result.headers['x-kss-request-id'])
#
# # list分块
# list_parts_params = {'uploadId': upload_id}
# list_parts_url = b.get_presigned_url(60, method="GET", params=list_parts_params, key_name=key)
# print(list_parts_url)
# result = requests.get(url=list_parts_url)
# print(result)
# print(result.content)
# print(result.headers['x-kss-request-id'])
# # 分块上传合并
# # complete_data = '''<?xml version="1.0" encoding="UTF-8"?>
# # <CompleteMultipartUpload>
# #   <Part>
# #     <PartNumber>1</PartNumber>
# #   </Part>
# # </CompleteMultipartUpload>'''
# # complete_mpu_url = b.get_presigned_url(60, method="POST", key_name=key, params={'uploadId':upload_id})
# # print(complete_mpu_url)
# # result = requests.post(url=complete_mpu_url, data=complete_data)
# # print(result)
# # print(result.content)
# # print(result.headers['x-kss-request-id'])
#
#
# # 终止分块上传
# abort_mpu_url = b.get_presigned_url(60, method="DELETE", key_name='DSC03380.mp.jpg', params={'uploadId': upload_id})
# print(abort_mpu_url)
# result = requests.delete(url=abort_mpu_url)
# print(result)
# print(result.content)
# print(result.headers['x-kss-request-id'])
# getAllBuckets()
# createBucket('')
# b = conn.get_bucket('test-bucket')
# key = b.get_key('hello')
# # 循环100次获取对象元数据
# for i in range(10):
#   resp = key.get_contents_as_string()
#   print(resp)
# print(newkey.size)
# uploadObjectFromString()
# newkey.set_contents_from_string('123456')




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
restore_object 方法使用示例

该示例演示了如何使用 restore_object 方法恢复归档的对象，
包括使用新增的 JobParameters 和 Tier 参数。
"""

import os
from ks3.connection import Connection

# 从环境变量获取配置
access_key = os.environ.get('KS3_IAM_ACCESS_KEY_ID')
secret_key = os.environ.get('KS3_IAM_ACCESS_KEY_SECRET')
bucket_name = os.environ.get('KS3_BUCKET')

# 初始化连接
conn = Connection(access_key, secret_key, host='ks3-cn-beijing.ksyuncs.com')
bucket = conn.get_bucket(bucket_name)


def example_restore_with_days_only():
    """示例1: 仅使用 days 参数恢复对象"""
    print("\n示例1: 仅使用 days 参数恢复对象")
    print("-" * 50)

    key_name = 'archived-file.txt'
    days = 3  # 恢复后保留3天

    try:
        result = bucket.restore_object(key_name, days=days)
        print(f"✓ 成功发起恢复请求")
        print(f"  状态码: {result.status}")
        print(f"  对象: {key_name}")
        print(f"  保留天数: {days}")
    except Exception as e:
        print(f"✗ 恢复失败: {e}")


def example_restore_with_tier_only():
    """示例2: 仅使用 tier 参数恢复对象"""
    print("\n示例2: 仅使用 tier 参数恢复对象")
    print("-" * 50)

    key_name = 'archived-file.txt'
    tier = 'Standard'  # 标准检索速度

    try:
        result = bucket.restore_object(key_name, tier=tier)
        print(f"✓ 成功发起恢复请求")
        print(f"  状态码: {result.status}")
        print(f"  对象: {key_name}")
        print(f"  检索层级: {tier}")
    except Exception as e:
        print(f"✗ 恢复失败: {e}")


def example_restore_with_days_and_tier():
    """示例3: 同时使用 days 和 tier 参数恢复对象"""
    print("\n示例3: 同时使用 days 和 tier 参数恢复对象")
    print("-" * 50)

    key_name = 'archived-file.txt'
    days = 5
    tier = 'Standard'

    try:
        result = bucket.restore_object(key_name, days=days, tier=tier)
        print(f"✓ 成功发起恢复请求")
        print(f"  状态码: {result.status}")
        print(f"  对象: {key_name}")
        print(f"  保留天数: {days}")
        print(f"  检索层级: {tier}")
    except Exception as e:
        print(f"✗ 恢复失败: {e}")


def example_restore_different_tiers():
    """示例4: 使用不同的检索层级"""
    print("\n示例4: 使用不同的检索层级")
    print("-" * 50)

    key_name = 'archived-file.txt'
    tiers = {
        'Expedited': '加急检索 - 1-5分钟内完成',
        'Standard': '标准检索 - 3-5小时内完成',
        'Bulk': '批量检索 - 5-12小时内完成'
    }

    for tier, description in tiers.items():
        print(f"\n使用 {tier} 层级:")
        print(f"  说明: {description}")
        try:
            result = bucket.restore_object(key_name, days=3, tier=tier)
            print(f"  ✓ 恢复请求已发起")
        except Exception as e:
            print(f"  ✗ 失败: {e}")


def example_restore_using_key_object():
    """示例5: 使用 Key 对象的 restore_object 方法"""
    print("\n示例5: 使用 Key 对象的 restore_object 方法")
    print("-" * 50)

    key_name = 'archived-file.txt'
    key = bucket.get_key(key_name)

    if key:
        try:
            result = key.restore_object(days=3, tier='Standard')
            print(f"✓ 成功发起恢复请求")
            print(f"  状态码: {result.status}")
            print(f"  对象: {key.name}")
        except Exception as e:
            print(f"✗ 恢复失败: {e}")
    else:
        print(f"✗ 对象 {key_name} 不存在")


def example_restore_with_custom_headers():
    """示例6: 使用自定义 headers"""
    print("\n示例6: 使用自定义 headers")
    print("-" * 50)

    key_name = 'archived-file.txt'
    headers = {
        'x-custom-header': 'custom-value'
    }

    try:
        result = bucket.restore_object(
            key_name,
            days=3,
            tier='Standard',
            headers=headers
        )
        print(f"✓ 成功发起恢复请求 (带自定义 headers)")
        print(f"  状态码: {result.status}")
    except Exception as e:
        print(f"✗ 恢复失败: {e}")


def example_restore_xml_format():
    """示例7: 生成的 XML 请求体格式说明"""
    print("\n示例7: 生成的 XML 请求体格式")
    print("-" * 50)

    print("\n当同时指定 days 和 tier 时，生成的 XML 格式如下:")
    print("""
<RestoreRequest>
    <Days>3</Days>
    <JobParameters>
        <Tier>Standard</Tier>
    </JobParameters>
</RestoreRequest>
    """)

    print("\n当仅指定 days 时:")
    print("""
<RestoreRequest>
    <Days>3</Days>
</RestoreRequest>
    """)

    print("\n当仅指定 tier 时:")
    print("""
<RestoreRequest>
    <JobParameters>
        <Tier>Standard</Tier>
    </JobParameters>
</RestoreRequest>
    """)


if __name__ == '__main__':
    print("=" * 50)
    print("restore_object 方法使用示例")
    print("=" * 50)

    # 注意: 以下示例需要实际的归档对象才能正常工作
    # 如果对象不存在或未归档，调用会返回错误

    # 取消下面的注释来运行示例
    # example_restore_with_days_only()
    # example_restore_with_tier_only()
    # example_restore_with_days_and_tier()
    # example_restore_different_tiers()
    # example_restore_using_key_object()
    # example_restore_with_custom_headers()

    # 显示 XML 格式说明（无需实际对象）
    example_restore_xml_format()

    print("\n" + "=" * 50)
    print("示例运行完毕")
    print("=" * 50)



# coding:utf-8
import base64
import json
import logging
import os
import time
import xml.sax

import requests

import ks3
from ks3 import auth
from ks3 import handler
from ks3 import utils
from ks3.auth import url_encode, format_policy_param
from ks3.bucket import Bucket, BucketLocation
from ks3.exception import S3ResponseError, S3CreateError, KS3ClientError, ParamValidationError
from ks3.http import make_request
from ks3.http import KS3HTTPAdapter
from ks3.provider import Provider
from ks3.responseResult import ResponseResult
from ks3.resultset import ResultSet
from ks3.utils import Ks3DnsResolver, RetryPolicy, is_bucket_name_valid
from ks3.xmlParsers.VpcAccessBlock import VpcAccessBlock
from ks3.xmlParsers.publicNetworkBlock import PublicNetworkBlock, BucketPublicNetworkBlock
from ks3.xmlParsers.batchJob import (
    CreateJobRequest, CreateJobResult, DescribeJobResult, ListJobsResult, UpdateJobPriorityResult,
    KS3PutObjectDataRedundancyTransition, KS3PutObjectAcl, KS3AccessControlList, KS3Grant,
    JobManifest, JobManifestFilter,
    JobManifestLocation, JobManifestSpec, JobOperation, JobReport
)
logger = logging.getLogger(__name__)

try:
    import urllib.parse as parse  # for Python 3
except ImportError:
    import urllib as parse  # for Python 2

dns_resolver = Ks3DnsResolver()

class _CallingFormat(object):

    def get_bucket_server(self, server, bucket):
        return ''

    def build_url_base(self, connection, protocol, server, bucket, key=''):
        url_base = '%s://' % protocol
        url_base += self.build_host(server, bucket)
        url_base += connection.get_path(self.build_path_base(bucket, key))
        return url_base

    def build_host(self, server, bucket):
        if bucket == '':
            return server
        else:
            return self.get_bucket_server(server, bucket)

    def build_auth_path(self, bucket, key=''):
        key = utils.get_utf8_value(key)
        path = ''
        if bucket != '':
            path = '/' + bucket
        buf = path + '/%s' % parse.quote(key)
        buf = buf.replace('//', '/%2F')
        return buf

    def build_path_base(self, bucket, key=''):
        buf = '/%s' % url_encode(key)
        buf = buf.replace('//', '/%2F')
        return buf


class PathCallingFormat(_CallingFormat):

    def get_bucket_server(self, server, bucket):
        return server

    def build_path_base(self, bucket, key=''):
        path_base = '/'
        if bucket:
            path_base += "%s/" % bucket
        path_base += url_encode(key)
        path_base = path_base.replace('//', '/%2F')
        return path_base


class SubdomainCallingFormat(_CallingFormat):

    def get_bucket_server(self, server, bucket):
        return '%s.%s' % (bucket, server)


# 习惯上倾向于称之为VirtualHostCallingFormat，故添加此别名类
class VirtualHostCallingFormat(SubdomainCallingFormat):
    pass


def _format_policy(expiration=None, conditions=[]):
  # if not expiration:
  #   raise ValueError("expiration must be set")
  policy = {
    'conditions': conditions
  }
  if expiration:
    policy['expiration'] = expiration

  logger.debug("policy: %s" % (policy))
  return policy


class Connection(object):
    QueryString = 'Signature=%s&Expires=%d&KSSAccessKeyId=%s'

    def __init__(self, access_key_id, access_key_secret, host="",
                 port=80, provider='kss', security_token=None, profile_name=None, path='/',
                 is_secure=False, debug=0, calling_format=VirtualHostCallingFormat(), domain_mode=False,
                 local_encrypt=False, local_key_path="", timeout=10, ua_addon='', enable_crc=True, block_size=8192,
                 proxy_host=None, proxy_port=None, proxy_username=None, proxy_password=None, connection_pool_size=1024,
                 dns_cache_time=30, retry_policy=RetryPolicy(), need_auth_header=True):
        """
        :param access_key_id: 金山云提供的ACCESS KEY ID
        :param access_key_secret: 金山云提供的SECRET KEY ID
        :param host: 请参考官网API文档说明中的Region定义(https://docs.ksyun.com/read/latest/65/_book/index.html)
        :param port: 请求端口，默认80
        :param is_secure: 是否启用HTTPS，True:启用  False:关闭
        :param domain_mode: 是否使用自定义域名访问，True:是 False:否
        :param local_encrypt: 是否启用本地加密， True:是 False:否，默认False，如选是，需要配置本地密钥路径
        :param enable_crc: 是否启用crc64校验，True:是 False:否，默认True
        :param block_size: 发送请求时，从本地读数据或者向服务器发数据的数据块大小
        """
        if host is None or host.strip() == '':
            raise ParamValidationError("host cannot be blank")

        self.access_key_id = access_key_id
        self.access_key_secret = access_key_secret
        self.is_secure = is_secure
        self.host = host
        self.port = port
        self.debug = debug
        self.path = path
        self.calling_format = calling_format
        self.domain_mode = domain_mode
        self.local_encrypt = local_encrypt
        self.key = ""
        self.timeout = timeout
        self.block_size = block_size
        self.ua_addon = ua_addon
        self.enable_crc = enable_crc
        self.need_auth_header = need_auth_header
        if self.is_secure:
            self.protocol = 'https'
            if self.port == 80:
                self.port = 443
        else:
            self.protocol = 'http'

        if isinstance(provider, Provider):
            # Allow overriding Provider
            self.provider = provider
        else:
            self._provider_type = provider
            self.provider = Provider(self._provider_type,
                                     access_key_id,
                                     access_key_secret,
                                     security_token,
                                     profile_name)

        # Allow config file to override default host, port, and host header.
        if self.provider.host:
            self.host = self.provider.host
        if self.provider.port:
            self.port = self.provider.port
        if self.provider.host_header:
            self.host_header = self.provider.host_header
        if self.local_encrypt:
            self.load_key(local_key_path)
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.proxy_username = proxy_username
        self.proxy_password = proxy_password

        self.session = requests.Session()
        adapter_kwargs = {
            'pool_connections': connection_pool_size,
            'pool_maxsize': connection_pool_size,
            'max_retries': 0,
            'block_size': block_size,
        }
        self.session.mount('http://', KS3HTTPAdapter(**adapter_kwargs))
        self.session.mount('https://', KS3HTTPAdapter(**adapter_kwargs))

        if dns_cache_time > 0 and dns_cache_time is not None:
            dns_resolver.ttl = dns_cache_time
            dns_resolver.open_cache()
        self.retry_policy = retry_policy

    def load_key(self, path):
        error_msg = "In local_encrypt mode, we need you to indicate the location of your private key. Set value for 'local_key_path' while initiate connection."
        assert path, error_msg
        with open(path, 'rb') as ak_file:
            assert os.path.getsize(path), "The key file should not be empty"
            content = ak_file.read()
            assert len(content.strip()) == 16, "The key's length should be 16"
            self.key = content.strip()

    def make_request(
            self, method, bucket="", key="", data="",
            headers=None, query_args=None, metadata=None, timeout=10,
            retry_policy=None,  # 兼容list方法的重试策略
    ):
        if not headers:
            headers = {}
        if not query_args:
            query_args = {}
        if not metadata:
            metadata = {}
        timeout = self.timeout

        request_kwargs = {
            'server': self.host,
            'port': self.port,
            'access_key_id': self.access_key_id,
            'access_key_secret': self.access_key_secret,
            'bucket': bucket,
            'key': key,
            'query_args': query_args,
            'headers': headers,
            'data': data,
            'metadata': metadata,
            'method': method,
            'is_secure': self.is_secure,
            'domain_mode': self.domain_mode,
            'timeout': timeout,
            'ua_addon': self.ua_addon,
            'proxy_host': self.proxy_host,
            'proxy_port': self.proxy_port,
            'proxy_username': self.proxy_username,
            'proxy_password': self.proxy_password,
            'calling_format': self.calling_format,
            'session': self.session,
            'need_auth_header': self.need_auth_header,
        }
        if retry_policy is None:
            retry_policy = self.retry_policy
        if retry_policy is not None:
            resp = retry_policy.call(
                make_request,
                **request_kwargs
            )
        else:
            resp = make_request(**request_kwargs)

        return resp

    def set_vpc_access_block(self, vpc_access_block):
        vab_xml = vpc_access_block.to_xml()
        if not isinstance(vab_xml, bytes):
            vab_xml = vab_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(vab_xml)
        headers = {'Content-MD5': md5, 'Content-Type': 'application/xml'}
        response = self.make_request('PUT', data=vab_xml, headers=headers, query_args='VpcAccessBlock')
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_vpc_access_block(self):
        response = self.make_request('GET', query_args='VpcAccessBlock')
        body = response.read()
        if response.status == 200:
            bta = VpcAccessBlock(status=response.status, reason=response.reason, headers=response.headers,
                                             raw_body=body)
            h = handler.XmlHandler(bta, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return bta
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_vpc_access_block(self):
        response = self.make_request('DELETE', query_args='VpcAccessBlock')
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    # Public Network Block methods
    def set_public_network_block(self, public_network_block):
        pnb_xml = public_network_block.to_xml()
        if not isinstance(pnb_xml, bytes):
            pnb_xml = pnb_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(pnb_xml)
        headers = {'Content-MD5': md5, 'Content-Type': 'application/xml'}
        response = self.make_request('PUT', data=pnb_xml, headers=headers, query_args='PublicNetworkBlock')
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_public_network_block(self):
        response = self.make_request('GET', query_args='PublicNetworkBlock')
        body = response.read()
        if response.status == 200:
            pnb = PublicNetworkBlock(status=response.status, reason=response.reason, headers=response.headers,
                                     raw_body=body)
            h = handler.XmlHandler(pnb, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return pnb
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_public_network_block(self):
        response = self.make_request('DELETE', query_args='PublicNetworkBlock')
        body = response.read()
        if response.status == 204 or response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_public_network_block(self, bucket_name, public_network_block):
        pnb_xml = public_network_block.to_xml()
        if not isinstance(pnb_xml, bytes):
            pnb_xml = pnb_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(pnb_xml)
        headers = {'Content-MD5': md5, 'Content-Type': 'application/xml'}
        response = self.make_request('PUT', bucket_name, data=pnb_xml, headers=headers, query_args='BucketPublicNetworkBlock')
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_public_network_block(self, bucket_name):
        response = self.make_request('GET', bucket_name, query_args='BucketPublicNetworkBlock')
        body = response.read()
        if response.status == 200:
            pnb = BucketPublicNetworkBlock(status=response.status, reason=response.reason, headers=response.headers,
                                           raw_body=body)
            h = handler.XmlHandler(pnb, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return pnb
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_public_network_block(self, bucket_name):
        response = self.make_request('DELETE', bucket_name, query_args='BucketPublicNetworkBlock')
        body = response.read()
        if response.status == 204 or response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_all_buckets(self, headers=None, project_ids=None,
                        region=None, prefix=None, visit_type=None, storage_type=None):
        query_args = {}
        if project_ids is not None:
            query_args = {
                "projectIds": project_ids
            }
        response = self.make_request('GET', headers=headers, query_args=query_args)
        body = response.read()
        if response.status > 300:
            raise S3ResponseError(response.status, response.reason, body)
        result = ResultSet(marker_elem=[('Bucket', Bucket)], status=response.status, reason=response.reason,
                           headers=response.headers, raw_body=body)
        h = handler.XmlHandler(result, self)
        if not isinstance(body, bytes):
            body = body.encode('utf-8')
        xml.sax.parseString(body, h)

        # 按条件过滤
        if region is not None:
            new_result = list(filter(lambda b: b.region == region, result))
            result.clear()
            result.extend(new_result)
        if prefix is not None:
            new_result = list(filter(lambda b: b.name.startswith(prefix), result))
            result.clear()
            result.extend(new_result)
        if visit_type is not None:
            new_result = list(filter(lambda b: b.VisitType == visit_type, result))
            result.clear()
            result.extend(new_result)
        if storage_type is not None:
            new_result = list(filter(lambda b: b.type == storage_type, result))
            result.clear()
            result.extend(new_result)

        return result

    def get_bucket(self, bucket_name, headers=None):
        return Bucket(self, bucket_name)

    def head_bucket(self, bucket_name, headers=None):
        response = self.make_request('HEAD', bucket_name)
        body = response.read()
        if response.status == 200:
            return ResponseResult(response.headers, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, None,
                                  request_id=response.getheader('x-kss-request-id'))

    def get_bucket_location(self, bucket_name):
        response = self.make_request('GET', bucket_name, query_args='location')
        body = response.read()
        if response.status == 200:
            loc = BucketLocation(status=response.status, reason=response.reason, headers=response.headers,
                                 raw_body=body)
            h = handler.XmlHandler(loc, self)
            xml.sax.parseString(body, h)
            return loc
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def create_bucket(self, bucket_name, headers=None, project_id=None,
                      location=None, policy=None, data_redundancy_type=None):
        is_bucket_name_valid(bucket_name)

        if policy:
            if headers:
                headers[self.provider.acl_header] = policy
            else:
                headers = {self.provider.acl_header: policy}

        data = None
        if location is not None:
            data = '<LocationConstraint>%s</LocationConstraint>'.format(location)
        if data_redundancy_type is not None:
            data = '<DataRedundancyType>%s</DataRedundancyType>'.format(data_redundancy_type)
        if data is not None:
            data = '<CreateBucketConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/">%s</CreateBucketConfiguration>'.format(data)
        else:
            data = ''

        query_args = {}
        if project_id is not None:
            query_args = {
                "projectId": project_id
            }
        response = self.make_request('PUT', bucket_name, headers=headers, query_args=query_args,
                                     data=data)
        body = response.read()
        if response.status == 409:
            raise S3CreateError(response.status, response.reason, body)
        if response.status == 200:
            return Bucket(self, bucket_name, status=response.status, reason=response.reason,
                          headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket(self, bucket_name, headers=None):
        """
        Removes an S3 bucket.

        In order to remove the bucket, it must first be empty. If the bucket is
        not empty, an ``S3ResponseError`` will be raised.
        """
        response = self.make_request('DELETE', bucket_name, headers=headers)
        body = response.read()
        if response.status != 204:
            raise S3ResponseError(response.status, response.reason, body)
        return ResponseResult(None, status=response.status, reason=response.reason,
                              headers=response.headers)

    def generate_url(self, expires_in, method, bucket='', key='', headers=None,
                     query_auth=True, force_http=False, response_headers=None, params=None,
                     expires_in_absolute=False, version_id=None, conditions=None):

        headers = headers or {}
        if expires_in_absolute:
          expires = int(expires_in)
        else:
          expires = int(time.time() + expires_in)
        extra_qp = []
        extra_qp_encoded = []
        query_args = None
        if version_id is not None:
          extra_qp.append("versionId=%s" % version_id)

        # response_headers应该被替换为params，
        # 但是为了兼容老版本，暂时保留response_headers
        # params更具备通用性，可以传入更多的参数
        all_params = params or {}
        if response_headers:
          logger.warning("""response_headers is deprecated. Please use params instead.""")
          all_params.update(response_headers)

        policy = None
        if conditions:
          policy = base64.b64encode(
            json.dumps(_format_policy(conditions=conditions)).encode('utf-8')).decode('utf-8')

        if all_params:
          for k, v in list(all_params.items()):
            if v is not None:
              extra_qp.append("%s=%s" % (k, v))
              extra_qp_encoded.append("%s=%s" % (k, parse.quote(str(v))))
            else:
              extra_qp.append("%s" % k)
              extra_qp_encoded.append("%s" % k)

        # 放入到query_args中，用来计算签名
        if extra_qp:
          query_args = '&'.join(extra_qp)

        # if not headers.has_key('Date'):
        #    headers['Date'] = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())

        c_string = auth.canonical_string(method, bucket, key, query_args=query_args, headers=headers, expires=expires, policy=policy)
        logger.debug('sign_str: {0}'.format(c_string))
        b64_hmac = auth.encode(self.access_key_secret, c_string)
        logger.debug('sign str encoded: {0}'.format(b64_hmac))
        encoded_canonical = parse.quote(b64_hmac, safe='')
        if query_auth:
          encode_ak = self.access_key_id
          # encode_ak = parse.quote(self.access_key_id)
          # print 'encode_ak:%s'%encode_ak
          query_part = '?' + self.QueryString % (encoded_canonical, expires, encode_ak)
          if policy:
            query_part += '&%s' % format_policy_param(policy)
        else:
          query_part = ''

        # 这段代码现在没有用
        if headers:
          hdr_prefix = self.provider.header_prefix
          for k, v in list(headers.items()):
            if k.startswith(hdr_prefix):
              # headers used for sig generation must be
              # included in the url also.
              extra_qp.append("%s=%s" % (k, parse.quote(v)))
        if extra_qp_encoded:
          delimiter = '?' if not query_part else '&'
          query_part += delimiter + '&'.join(extra_qp_encoded)
        if force_http:
          protocol = 'http'
          port = 80
        else:
          protocol = self.protocol
          port = self.port
        return self.calling_format.build_url_base(self, protocol,
                                                  self.server_name(port),
                                                  bucket, key) + query_part

    def server_name(self, port=None):
        if not port:
            port = self.port
        if port == 80:
            signature_host = self.host
        else:
            signature_host = "%s:%s" % (self.host, port)
        return signature_host

    def get_path(self, path='/'):
        # The default behavior is to suppress consecutive slashes for reasons
        # discussed at
        # https://groups.google.com/forum/#!topic/boto-dev/-ft0XPUy0y8
        # You can override that behavior with the suppress_consec_slashes param.
        pos = path.find('?')
        if pos >= 0:
            params = path[pos:]
            path = path[:pos]
        else:
            params = None
        if path[-1] == '/':
            need_trailing = True
        else:
            need_trailing = False
        path_elements = self.path.split('/')
        path_elements.extend(path.split('/'))
        path_elements = [p for p in path_elements if p]
        path = '/' + '/'.join(path_elements)
        if path[-1] != '/' and need_trailing:
            path += '/'
        if params:
            path = path + params
        return path

    def get_adp(self, task_id):
        query_args = 'queryadp'
        response = self.make_request('GET', task_id, query_args=query_args)
        body = response.read()
        if response.status != 200:
          raise S3ResponseError(response.status, response.reason, body)
        return ResponseResult(body, status=response.status, reason=response.reason,
                              headers=response.headers)

    def create_job(self, manifest, operation, priority=None, description=None,
                   report=None, client_request_token=None):
        """
        创建批量任务

        :param manifest: 任务清单，包含需要处理的对象列表
        :param operation: 任务操作类型
        :param priority: 任务优先级（可选）
        :param description: 任务描述（可选）
        :param report: 任务报告配置（可选）
        :param client_request_token: 客户端请求令牌（可选）

        :rtype: CreateJobResult
        :return: 创建任务的结果，包含任务ID
        """
        request = CreateJobRequest(
            manifest=manifest,
            operation=operation,
            priority=priority,
            description=description,
            report=report,
            client_request_token=client_request_token
        )
        data = request.to_xml()

        if not isinstance(data, bytes):
            data = data.encode('utf-8')
        headers = {'Content-Type': 'application/xml'}
        md5 = ks3.utils.compute_base64_md5_digest(data)
        headers['Content-MD5'] = md5

        response = self.make_request(
            'PUT', data=data,
            headers=headers,
            query_args='jobs'
        )
        body = response.read()
        if response.status == 200:
            result = CreateJobResult(status=response.status, reason=response.reason,
                                     headers=response.headers, raw_body=body)
            h = handler.XmlHandler(result, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return result
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def describe_job(self, job_id):
        """
        查询任务详情

        :param job_id: 任务ID

        :rtype: DescribeJobResult
        :return: 任务详情
        """
        query_args = 'jobs&jobId=%s' % job_id
        response = self.make_request('GET', query_args=query_args)
        body = response.read()
        if response.status == 200:
            result = DescribeJobResult(status=response.status, reason=response.reason,
                                        headers=response.headers, raw_body=body)
            h = handler.XmlHandler(result, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return result
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_job(self, job_id):
        """
        删除任务

        :param job_id: 任务ID

        :rtype: ResponseResult
        :return: 删除结果
        """
        query_args = 'jobs&jobId=%s' % job_id
        response = self.make_request('DELETE', query_args=query_args)
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def list_jobs(self, next_token=None, max_results=None, job_statuses=None):
        """
        列出任务列表

        :param next_token: 分页令牌（可选）
        :param max_results: 最大返回数量（可选）
        :param job_statuses: 任务状态过滤（可选）

        :rtype: ListJobsResult
        :return: 任务列表
        """
        query_args = ['jobs']
        if next_token is not None:
            query_args.append('nextToken=%s' % next_token)
        if max_results is not None:
            query_args.append('maxResults=%s' % max_results)
        if job_statuses is not None:
            query_args.append('jobStatuses=%s' % job_statuses)
        query_args = '&'.join(query_args)
        response = self.make_request('GET', query_args=query_args)
        body = response.read()
        if response.status == 200:
            result = ListJobsResult(status=response.status, reason=response.reason,
                                     headers=response.headers, raw_body=body)
            h = handler.XmlHandler(result, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return result
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def update_job_priority(self, job_id, priority):
        """
        更新任务优先级

        :param job_id: 任务ID
        :param priority: 新的优先级

        :rtype: UpdateJobPriorityResult
        :return: 更新结果
        """
        query_args = 'jobs&jobId=%s&action=updateJobPriority&priority=%s' % (job_id, priority)
        response = self.make_request('PUT', query_args=query_args)
        body = response.read()
        if response.status == 200:
            result = UpdateJobPriorityResult(status=response.status, reason=response.reason,
                                             headers=response.headers, raw_body=body)
            h = handler.XmlHandler(result, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return result
        else:
            raise S3ResponseError(response.status, response.reason, body)


# Copyright (c) 2006,2007 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import xml.sax

from .compat import StringIO


class XmlHandler(xml.sax.ContentHandler):

    def __init__(self, root_node, connection):
        self.connection = connection
        self.nodes = [('root', root_node)]
        self.current_text = ''

    def startElement(self, name, attrs):
        self.current_text = ''
        new_node = self.nodes[-1][1].startElement(name, attrs, self.connection)
        if new_node is not None:
            self.nodes.append((name, new_node))

    def endElement(self, name):
        self.nodes[-1][1].endElement(name, self.current_text, self.connection)
        if self.nodes[-1][0] == name:
            if hasattr(self.nodes[-1][1], 'endNode'):
                self.nodes[-1][1].endNode(self.connection)
            self.nodes.pop()
        self.current_text = ''

    def characters(self, content):
        self.current_text += content


class XmlHandlerWrapper(object):
    def __init__(self, root_node, connection):
        self.handler = XmlHandler(root_node, connection)
        self.parser = xml.sax.make_parser()
        self.parser.setContentHandler(self.handler)
        self.parser.setFeature(xml.sax.handler.feature_external_ges, 0)

    def parseString(self, content):
        return self.parser.parse(StringIO(content))
